import re
import sys
from pygls.lsp.types.language_features import Location, Position, Range
from typing import List, Optional

from pygls.workspace import Document
from vyper_lsp.ast import AST
from vyper_lsp.utils import get_expression_at_cursor, get_word_at_cursor


# this class should abstract away all the AST stuff
# and just provide a simple interface for navigation
#
# the navigator should mainly return Ranges
class ASTNavigator:
    def __init__(self, ast=None):
        self.ast = ast or AST()

    def find_state_variable_declaration(self, word: str) -> Optional[Range]:
        node = self.ast.find_state_variable_declaration_node(word)
        if node:
            range = Range(
                start=Position(line=node.lineno - 1, character=node.col_offset),
                end=Position(line=node.end_lineno - 1, character=node.end_col_offset),
            )
            return range

    def find_function_declaration(self, word: str) -> Optional[Range]:
        node = self.ast.find_function_declaration_node_for_name(word)
        if node:
            range = Range(
                start=Position(line=node.lineno - 1, character=node.col_offset),
                end=Position(line=node.end_lineno - 1, character=node.end_col_offset),
            )
            return range

    def find_type_declaration(self, word: str) -> Optional[Range]:
        node = self.ast.find_type_declaration_node(word)
        if node:
            range = Range(
                start=Position(line=node.lineno - 1, character=node.col_offset),
                end=Position(line=node.end_lineno - 1, character=node.end_col_offset),
            )
            return range

    def find_references(self, doc: Document, pos: Position) -> List[Range]:
        og_line = doc.lines[pos.line]
        word = get_word_at_cursor(og_line, pos.character)
        if self.ast.ast_data is None:
            return []
        references = []

        if word in self.ast.get_enums():
            # find all references to this type
            refs = self.ast.find_nodes_referencing_enum(word)
            for ref in refs:
                range = Range(
                    start=Position(line=ref.lineno - 1, character=ref.col_offset),
                    end=Position(line=ref.end_lineno - 1, character=ref.end_col_offset),
                )
                references.append(range)
        elif word in self.ast.get_structs() or word in self.ast.get_events():
            refs = self.ast.find_nodes_referencing_struct(word)
            for ref in refs:
                range = Range(
                    start=Position(line=ref.lineno - 1, character=ref.col_offset),
                    end=Position(line=ref.end_lineno - 1, character=ref.end_col_offset),
                )
                references.append(range)
        elif not og_line[0].isspace() and word in self.ast.get_state_variables():
            if "constant(" in og_line:
                refs = self.ast.find_nodes_referencing_constant(word)
            else:
                refs = self.ast.find_nodes_referencing_state_variable(word)
            for ref in refs:
                range = Range(
                    start=Position(line=ref.lineno - 1, character=ref.col_offset),
                    end=Position(line=ref.end_lineno - 1, character=ref.end_col_offset),
                )
                references.append(range)
        return references

    def find_declaration(self, document: Document, pos: Position) -> Optional[Range]:
        if self.ast.ast_data is None:
            return None

        og_line = document.lines[pos.line]
        word = get_word_at_cursor(og_line, pos.character)
        full_word = get_expression_at_cursor(og_line, pos.character)
        range = None

        print(f"word: {word}", file=sys.stderr)

        if full_word.startswith("self."):
            if "(" in full_word:
                range = self.find_function_declaration(word)
            else:
                range = self.find_state_variable_declaration(word)
        else:
            print(f"user defined types: {self.ast.get_user_defined_types()}", file=sys.stderr)
            if word in self.ast.get_user_defined_types():
                range = self.find_type_declaration(word)
            elif word in self.ast.get_constants():
                range = self.find_state_variable_declaration(word)
            else:
                # check if full_word matches "enum.variant" regex
                pattern = r"([a-zA-Z_][a-zA-Z0-9_]*)\.([a-zA-Z_][a-zA-Z0-9_]*)"
                match = re.match(pattern, full_word)
                if match:
                    enum_name = match.group(1)
                    variant_name = match.group(2)
                    if enum_name in self.ast.get_enums():
                        variants = self.ast.get_enum_variants(enum_name)
                        if variant_name in variants:
                            range = self.find_type_declaration(enum_name)
        if range:
            return range

    def find_implementation(self, document: Document, pos: Position) -> Optional[Range]:
        og_line = document.lines[pos.line]
        word = get_word_at_cursor(og_line, pos.character)
        expression = get_expression_at_cursor(og_line, pos.character)

        if not "(" in expression:
            return None

        if expression.startswith("self."):
            return self.find_function_declaration(word)
        elif og_line[0].isspace() and og_line.strip().startswith("def"):
            # only lookup external fns if we're in an interface def
            return self.find_function_declaration(word)
        else:
            return None



import re
from pathlib import Path
from importlib.metadata import version
from packaging.version import Version

from vyper.compiler import CompilerData


def get_installed_vyper_version():
    return Version(version("vyper"))

def get_source(filepath):
    base_path = Path(__file__).parent.parent
    filepath = base_path / filepath
    return filepath.read_text()


def get_compiler_data(filepath):
    source = get_source(filepath)
    return CompilerData(source)


# detect if current line is a variable declaration
def is_var_declaration(line):
    # regex for variable declaration
    # should match lines starting with any identifier followed by a colon
    # like "foo: "
    reg = r"^\s*[a-zA-Z_][a-zA-Z0-9_]*\s*:"
    return bool(re.match(reg, line.strip()))


def is_attribute_access(line):
    # regex for attribute access
    # should match lines ending with a dot
    # like "foo."
    reg = r"\s*\.\s*$"
    return bool(re.match(reg, line.strip()))


def get_word_at_cursor(sentence: str, cursor_index: int) -> str:
    start = cursor_index
    end = cursor_index

    # Find the start of the word
    while start > 0 and sentence[start - 1].isalnum():
        start -= 1

    # Find the end of the word
    while end < len(sentence) and sentence[end].isalnum():
        end += 1

    # Extract the word
    word = sentence[start:end]

    return word


def get_expression_at_cursor(sentence: str, cursor_index: int) -> str:
    # does the same thing as get_word_at_cursor but includes . and [ and ] in the expression
    start = cursor_index
    end = cursor_index

    # Find the start of the word
    while start > 0 and sentence[start - 1].isalnum() or sentence[start - 1] in ".[]()":
        start -= 1

    # Find the end of the word
    while end < len(sentence) and sentence[end].isalnum() or sentence[end] in ".[]()":
        end += 1

    # Extract the word
    word = sentence[start:end]

    return word


def extract_enum_name(line: str):
    match = re.match(r"enum\s+([a-zA-Z_][a-zA-Z0-9_]*)\s*:", line)
    if match:
        return match.group(1)
    return None



import sys
from pygls.lsp.types.language_features import List
from vyper.ast import nodes
from vyper.compiler import CompilerData

ast = None


class AST:
    _instance = None
    ast_data = None

    custom_type_node_types = (nodes.StructDef, nodes.EnumDef, nodes.EventDef)

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(AST, cls).__new__(cls)
            cls._instance.ast_data = None
        return cls._instance

    def update_ast(self, document):
        try:
            compiler_data = CompilerData(document.source)
            self.ast_data = compiler_data.vyper_module_unfolded
        except Exception as e:
            print(f"Error generating unfolded AST, {e}")
            pass

    def build_ast(self, src: str):
        try:
            compiler_data = CompilerData(src)
            self.ast_data = compiler_data.vyper_module_unfolded
        except Exception as e:
            print(f"Error generating unfolded AST, {e}")
            pass

    def get_enums(self) -> List[str]:
        if self.ast_data is None:
            return []

        return [node.name for node in self.ast_data.get_descendants(nodes.EnumDef)]

    def get_structs(self) -> List[str]:
        if self.ast_data is None:
            return []

        return [node.name for node in self.ast_data.get_descendants(nodes.StructDef)]

    def get_events(self) -> List[str]:
        if self.ast_data is None:
            return []

        return [node.name for node in self.ast_data.get_descendants(nodes.EventDef)]

    def get_user_defined_types(self):
        if self.ast_data is None:
            return []

        return [
            node.name
            for node in self.ast_data.get_descendants(self.custom_type_node_types)
        ]

    def get_constants(self):
        if self.ast_data is None:
            return []

        return [
            node.target.id
            for node in self.ast_data.get_descendants(nodes.VariableDecl)
            if node.is_constant
        ]

    def get_enum_variants(self, enum: str):
        if self.ast_data is None:
            return []

        enum_node = self.find_type_declaration_node(enum)
        if enum_node is None:
            return []

        return [node.value.id for node in enum_node.get_children()]

    def get_state_variables(self):
        if self.ast_data is None:
            return []

        print(f"{self.ast_data.get_descendants(nodes.VariableDecl)}", file=sys.stderr)

        return [
            node.target.id for node in self.ast_data.get_descendants(nodes.VariableDecl)
        ]

    def get_internal_functions(self):
        if self.ast_data is None:
            return []

        function_nodes = self.ast_data.get_descendants(nodes.FunctionDef)
        inernal_nodes = []

        for node in function_nodes:
            for decorator in node.decorator_list:
                if decorator.id == "internal":
                    inernal_nodes.append(node)

        return inernal_nodes

    def find_nodes_referencing_internal_function(self, function: str):
        if self.ast_data is None:
            return []

        return self.ast_data.get_descendants(
            nodes.Call, {"func.attr": function, "func.value.id": "self"}
        )

    def find_nodes_referencing_state_variable(self, variable: str):
        if self.ast_data is None:
            return []

        return self.ast_data.get_descendants(
            nodes.Attribute, {"value.id": "self", "attr": variable}
        )

    def find_nodes_referencing_constant(self, constant: str):
        if self.ast_data is None:
            return []

        name_nodes = self.ast_data.get_descendants(nodes.Name, {"id": constant})
        return [
            node
            for node in name_nodes
            if not isinstance(node.get_ancestor(), nodes.VariableDecl)
        ]

    def get_attributes_for_symbol(self, symbol: str):
        if self.ast_data is None:
            return []

        node = self.find_type_declaration_node(symbol)
        if node is None:
            return []

        if isinstance(node, nodes.StructDef):
            return []
        elif isinstance(node, nodes.EnumDef):
            return self.get_enum_variants(symbol)
        else:
            return []

    def find_function_declaration_node_for_name(self, function: str):
        if self.ast_data is None:
            return None

        for node in self.ast_data.get_descendants(nodes.FunctionDef):
            name_match = node.name == function
            not_interface_declaration = not isinstance(
                node.get_ancestor(), nodes.InterfaceDef
            )
            if name_match and not_interface_declaration:
                return node

        return None

    def find_state_variable_declaration_node(self, variable: str):
        if self.ast_data is None:
            return None

        for node in self.ast_data.get_descendants(nodes.VariableDecl):
            if node.target.id == variable:
                return node

        return None

    def find_type_declaration_node(self, symbol: str):
        if self.ast_data is None:
            return None

        for node in self.ast_data.get_descendants(self.custom_type_node_types):
            if node.name == symbol:
                return node
            if isinstance(node, nodes.EnumDef):
                for variant in node.get_children(nodes.Expr):
                    if variant.value.id == symbol:
                        return variant

        return None

    def find_nodes_referencing_enum(self, enum: str):
        if self.ast_data is None:
            return []

        return_nodes = []

        for node in self.ast_data.get_descendants(
            nodes.AnnAssign, {"annotation.id": enum}
        ):
            return_nodes.append(node)
        for node in self.ast_data.get_descendants(nodes.Attribute, {"value.id": enum}):
            return_nodes.append(node)
        for node in self.ast_data.get_descendants(
            nodes.VariableDecl, {"annotation.id": enum}
        ):
            return_nodes.append(node)

        return return_nodes

    def find_nodes_referencing_enum_variant(self, enum: str, variant: str):
        if self.ast_data is None:
            return None

        return self.ast_data.get_descendants(
            nodes.Attribute, {"attr": variant, "value.id": enum}
        )

    def find_nodes_referencing_struct(self, struct: str):
        if self.ast_data is None:
            return []

        return_nodes = []

        for node in self.ast_data.get_descendants(
            nodes.AnnAssign, {"annotation.id": struct}
        ):
            return_nodes.append(node)
        for node in self.ast_data.get_descendants(nodes.Call, {"func.id": struct}):
            return_nodes.append(node)
        for node in self.ast_data.get_descendants(
            nodes.VariableDecl, {"annotation.id": struct}
        ):
            return_nodes.append(node)
        for node in self.ast_data.get_descendants(
            nodes.FunctionDef, {"returns.id": struct}
        ):
            return_nodes.append(node)

        return return_nodes



import sys
import re
from typing import List, Optional
import warnings
from packaging.version import Version
from pygls.lsp.types import Diagnostic, Position, Range
from pygls.workspace import Document
from vyper.compiler import CompilerData
from vyper.exceptions import VyperException
from vyper_lsp.analyzer.BaseAnalyzer import Analyzer
from vyper_lsp.ast import AST
from vyper_lsp.utils import get_expression_at_cursor, get_word_at_cursor, get_installed_vyper_version
from pygls.lsp.types.language_features import (
    CompletionItem,
    CompletionList,
    CompletionParams,
)
from pygls.server import LanguageServer

pattern = r"(.+) is deprecated\. Please use `(.+)` instead\."
compiled_pattern = re.compile(pattern)

min_vyper_version = Version("0.3.7")

# Available base types
UNSIGNED_INTEGER_TYPES = {f"uint{8*(i+1)}" for i in range(32)}
SIGNED_INTEGER_TYPES = {f"int{8*(i+1)}" for i in range(32)}
INTEGER_TYPES = UNSIGNED_INTEGER_TYPES | SIGNED_INTEGER_TYPES

BYTES_M_TYPES = {f"bytes{i+1}" for i in range(32)}
DECIMAL_TYPES = {"decimal"}

BASE_TYPES = INTEGER_TYPES | BYTES_M_TYPES | DECIMAL_TYPES | {"bool", "address"}

DECORATORS = ["payable", "nonpayable", "view", "pure", "external", "internal"]

class AstAnalyzer(Analyzer):

    def __init__(self, ast=None) -> None:
        super().__init__()
        self.ast = ast or AST()
        if get_installed_vyper_version() < min_vyper_version:
            self.diagnostics_enabled = False
        else:
            self.diagnostics_enabled = True

    def get_completions(
        self, ls: LanguageServer, params: CompletionParams
    ) -> CompletionList:
        items = []
        document = ls.workspace.get_document(params.text_document.uri)
        current_line = document.lines[params.position.line].strip()
        custom_types = self.ast.get_user_defined_types()

        if params.context:
            if params.context.trigger_character == ".":
                # get element before the dot
                element = current_line.split(" ")[-1].split(".")[0]
                for attr in self.ast.get_attributes_for_symbol(element):
                    items.append(CompletionItem(label=attr))
                l = CompletionList(is_incomplete=False, items=[])
                l.add_items(items)
                return l
            elif params.context.trigger_character == "@":
                for dec in DECORATORS:
                    items.append(CompletionItem(label=dec))
                l = CompletionList(is_incomplete=False, items=[])
                l.add_items(items)
                return l
            elif params.context.trigger_character == ":":
                for typ in custom_types + list(BASE_TYPES):
                    items.append(CompletionItem(label=typ, insert_text=f" {typ}"))

                l = CompletionList(is_incomplete=False, items=[])
                l.add_items(items)
                return l
            else:
                if params.context.trigger_character == " ":
                    if current_line[-1] == ":":
                        for typ in custom_types + list(BASE_TYPES):
                            items.append(CompletionItem(label=typ))

                        l = CompletionList(is_incomplete=False, items=[])
                        l.add_items(items)
                        return l
                return CompletionList(is_incomplete=False, items=[])

        else:
            return CompletionList(is_incomplete=False, items=[])

    def hover_info(self, document: Document, pos: Position) -> Optional[str]:
        og_line = document.lines[pos.line]
        word = get_word_at_cursor(og_line, pos.character)
        full_word = get_expression_at_cursor(og_line, pos.character)

        if full_word.startswith("self."):
            if "(" in full_word:
                node = self.ast.find_function_declaration_node_for_name(word)
                if node:
                    fn_name = node.name
                    arg_str = ", ".join(
                        [f"{arg.arg}: {arg.annotation.id}" for arg in node.args.args]
                    )
                    return f"(Internal Function) **{fn_name}**({arg_str})"
            else:
                node = self.ast.find_state_variable_declaration_node(word)
                if node:
                    variable_type = node.annotation.id
                    return f"(State Variable) **{word}** : **{variable_type}**"
        else:
            if word in self.ast.get_structs():
                node = self.ast.find_type_declaration_node(word)
                if node:
                    return f"(Struct) **{word}**"
            elif word in self.ast.get_enums():
                node = self.ast.find_type_declaration_node(word)
                if node:
                    return f"(Enum) **{word}**"
            elif word in self.ast.get_events():
                node = self.ast.find_type_declaration_node(word)
                if node:
                    return f"(Event) **{word}**"
            elif word in self.ast.get_constants():
                node = self.ast.find_state_variable_declaration_node(word)
                if node:
                    variable_type = node.annotation.id
                    return f"(Constant) **{word}** : **{variable_type}**"
            else:
                return None

    def get_diagnostics(self, doc: Document) -> List[Diagnostic]:
        diagnostics = []

        if not self.diagnostics_enabled:
            return diagnostics

        replacements = {}
        warnings.simplefilter("always")
        with warnings.catch_warnings(record=True) as w:
            try:
                compiler_data = CompilerData(doc.source)
                compiler_data.vyper_module_folded
            except VyperException as e:
                if e.lineno is not None and e.col_offset is not None:
                    diagnostics.append(
                        Diagnostic(
                            range=Range(
                                start=Position(
                                    line=e.lineno - 1, character=e.col_offset - 1
                                ),
                                end=Position(line=e.lineno - 1, character=e.col_offset),
                            ),
                            message=str(e),
                            severity=1,
                        )
                    )
                else:
                    for a in e.annotations:
                        diagnostics.append(
                            Diagnostic(
                                range=Range(
                                    start=Position(
                                        line=a.lineno - 1, character=a.col_offset - 1
                                    ),
                                    end=Position(
                                        line=a.lineno - 1, character=a.col_offset
                                    ),
                                ),
                                message=e.message,
                                severity=1,
                            )
                        )
            for warning in w:
                match = compiled_pattern.match(str(warning.message))
                if not match:
                    continue
                deprecated = match.group(1)
                replacement = match.group(2)
                replacements[deprecated] = replacement

        # iterate over doc.lines and find all deprecated values
        # and create a warning for each one at the correct position
        for i, line in enumerate(doc.lines):
            for deprecated, replacement in replacements.items():
                if deprecated in line:
                    diagnostics.append(
                        Diagnostic(
                            range=Range(
                                start=Position(
                                    line=i, character=line.index(deprecated)
                                ),
                                end=Position(
                                    line=i,
                                    character=line.index(deprecated) + len(deprecated),
                                ),
                            ),
                            message=f"{deprecated} is deprecated. Please use {replacement} instead.",
                            severity=2,
                        )
                    )

        return diagnostics



import re
import sys
from typing import List, Optional
from lark import UnexpectedInput, UnexpectedToken
from packaging.specifiers import Specifier
from packaging.version import Version
from pygls.lsp.types import CompletionList, CompletionParams, Diagnostic, Position, Range
from pygls.lsp.types.language_features import on_type_formatting
from pygls.server import LanguageServer
from pygls.workspace import Document
from vvm.exceptions import VyperError
from vyper.exceptions import VersionException
from vyper_lsp.analyzer.BaseAnalyzer import Analyzer
import vvm

from pathlib import Path

import vyper
from lark import Lark
from vyper.ast.grammar import PythonIndenter

GRAMMAR_FILE_PATH = Path(vyper.__file__).parent / "ast" / "grammar.lark"
GRAMMAR = GRAMMAR_FILE_PATH.read_text()

parser = Lark(GRAMMAR, parser="lalr", start="module", postlex=PythonIndenter())


def format_parse_error(e):
    if isinstance(e, UnexpectedToken):
        expected = ", ".join(e.accepts or e.expected)
        return f"Unexpected token '{e.token}' at {e.line}:{e.column}. Expected one of: {expected}"
    else:
        return str(e)

LEGACY_VERSION_PRAGMA_REGEX = re.compile(r"^#\s*@version\s+(.*)$")
VERSION_PRAGMA_REGEX = re.compile(r"^#pragma\s+version\s+(.*)$")

def extract_version_pragma(line: str) -> Optional[str]:
    if match := LEGACY_VERSION_PRAGMA_REGEX.match(line):
        return match.group(1)
    elif match := VERSION_PRAGMA_REGEX.match(line):
        return match.group(1)
    else:
        return None

class SourceAnalyzer(Analyzer):

    def __init__(self) -> None:
        self.parser_enabled = True
        self.compiler_enabled = False

    def get_version_pragma(self, doc: Document) -> Optional[str]:
        doc_lines = doc.lines
        for line in doc_lines:
            if version := extract_version_pragma(line):
                return version

    def hover_info(self, doc: Document, pos: Position) -> Optional[str]:
        return None

    def get_parser_diagnostics(self, doc: Document) -> List[Diagnostic]:
        diagnostics = []
        last_error = None

        def on_grammar_error(e: UnexpectedInput) -> bool:
            nonlocal last_error
            if (
                last_error is not None
                and last_error.line == e.line
                and last_error.column == e.column
            ):
                return True
            if (
                last_error is not None
                and type(last_error) == type(e)
                and last_error.line == e.line
            ):
                return True
            diagnostics.append(
                Diagnostic(
                    range=Range(
                        start=Position(line=e.line - 1, character=e.column - 1),
                        end=Position(line=e.line - 1, character=e.column),
                    ),
                    message=format_parse_error(e),
                )
            )
            last_error = e
            return True

        try:
            parser.parse(doc.source, on_error=on_grammar_error)
        except Exception:
            # ignore errors that are already handled by on_error
            pass

        return diagnostics

    def get_compiler_diagnostics(self, doc: Document) -> List[Diagnostic]:
        diagnostics = []

        # now compile via vvm to get semantic errors
        try:
            vvm.compile_source(doc.source)
        except VyperError as e:
            if "vyper.exceptions.VersionException" in e.stderr_data:
                # ignore version errors, we handle them separately
                version_pragma = self.get_version_pragma(doc)
                # check if version pragma is just a version or a specifier
                if version_pragma is None:
                    diagnostics.append(
                        Diagnostic(
                            range=Range(
                                start=Position(line=0, character=0),
                                end=Position(line=0, character=0),
                            ),
                            message=str(e),
                        )
                    )
                elif re.match(r"^\d+\.\d+\.\d+$", version_pragma):
                    # version pragma is just a version, so we just install it
                    version: Version = Version(version_pragma)
                    vvm.install_vyper(version)
                    vvm.set_vyper_version(version)
                else:
                    specifier: Specifier = Specifier(version_pragma)
                    for version in vvm.get_installable_vyper_versions():
                        if specifier.contains(version):
                            vvm.install_vyper(version)
                            vvm.set_vyper_version(version)
            else:
                # find the line and number of the error
                # will appear in the text as `line x:y`
                regex = re.compile(r"line (\d+):(\d+)")
                match = regex.search(e.stderr_data)

                type_regex = re.compile(r"vyper\.exceptions\.([a-zA-Z]+): (.*)\n")
                type_match = type_regex.search(e.stderr_data)
                if match and type_match:
                    diagnostics.append(
                        Diagnostic(
                            range=Range(
                                start=Position(
                                    line=int(match.group(1)) - 1,
                                    character=int(match.group(2)) - 1,
                                ),
                                end=Position(
                                    line=int(match.group(1)) - 1,
                                    character=int(match.group(2)),
                                ),
                            ),
                            message=f"{type_match.group(1)}: {type_match.group(2)}",
                        )
                    )
        except:
            print("got unknown exception", file=sys.stderr)

        return diagnostics

    def get_diagnostics(self, doc: Document) -> List[Diagnostic]:
        diagnostics = []
        if self.parser_enabled:
            diagnostics.extend(self.get_parser_diagnostics(doc))
        if self.compiler_enabled:
            diagnostics.extend(self.get_compiler_diagnostics(doc))
        return diagnostics

    def get_completions(self, ls: LanguageServer, params: CompletionParams) -> CompletionList:
        completions: CompletionList = CompletionList(is_incomplete=False, items=[])
        return completions
